@@ -3,7 +3,7 @@
  * tsvector_op.c
  *	  operations over tsvector
  *
- * Portions Copyright (c) 1996-2018, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2019, PostgreSQL Global Development Group
  *
  *
  * IDENTIFICATION
@@ -65,23 +65,8 @@ typedef struct
 	StatEntry  *root;
 } TSVectorStat;
 
-#define STATHDRSIZE (offsetof(TSVectorStat, data))
-
-/* TS_execute requires ternary logic to handle NOT with phrase matches */
-typedef enum
-{
-	TS_NO,						/* definitely no match */
-	TS_YES,						/* definitely does match */
-	TS_MAYBE					/* can't verify match for lack of pos data */
-} TSTernaryValue;
-
-
-static TSTernaryValue TS_execute_recurse(QueryItem *curitem, void *arg,
-										 uint32 flags,
-										 TSExecuteCallback chkcond);
-static int	tsvector_bsearch(const TSVector tsv, char *lexeme, int lexeme_len);
 static Datum tsvector_update_trigger(PG_FUNCTION_ARGS, bool config_column);
-
+static int	tsvector_bsearch(const TSVector tsv, char *lexeme, int lexeme_len);
 
 /*
  * Order: haspos, len, word, for all positions (pos, weight)
@@ -659,7 +644,7 @@ tsvector_unnest(PG_FUNCTION_ARGS)
 		funcctx = SRF_FIRSTCALL_INIT();
 		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
 
-		tupdesc = CreateTemplateTupleDesc(3, false);
+		tupdesc = CreateTemplateTupleDesc(3);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 1, "lexeme",
 						   TEXTOID, -1, 0);
 		TupleDescInitEntry(tupdesc, (AttrNumber) 2, "positions",
@@ -1327,13 +1312,12 @@ checkcondition_str(void *checkval, QueryOperand *val, ExecPhraseData *data)
 	WordEntry  *StopLow = chkval->arrb;
 	WordEntry  *StopHigh = chkval->arre;
 	WordEntry  *StopMiddle = StopHigh;
+	int			difference = -1;
 	bool		res = false;
 
 	/* Loop invariant: StopLow <= val < StopHigh */
 	while (StopLow < StopHigh)
 	{
-		int			difference;
-
 		StopMiddle = StopLow + (StopHigh - StopLow) / 2;
 		difference = tsCompareString(chkval->operand + val->distance,
 									 val->length,
@@ -1399,11 +1383,6 @@ checkcondition_str(void *checkval, QueryOperand *val, ExecPhraseData *data)
 					memcpy(allpos + npos, data->pos, sizeof(WordEntryPos) * data->npos);
 					npos += data->npos;
 				}
-				else
-				{
-					/* at loop exit, res must be true if we found matches */
-					res = (npos > 0);
-				}
 			}
 			else
 			{
@@ -1442,17 +1421,14 @@ checkcondition_str(void *checkval, QueryOperand *val, ExecPhraseData *data)
  * Loffset and Roffset should not be negative, else we risk trying to output
  * negative positions, which won't fit into WordEntryPos.
  *
- * The result is boolean (TS_YES or TS_NO), but for the caller's convenience
- * we return it as TSTernaryValue.
- *
- * Returns TS_YES if any positions were emitted to *data; or if data is NULL,
- * returns TS_YES if any positions would have been emitted.
+ * Returns true if any positions were emitted to *data; or if data is NULL,
+ * returns true if any positions would have been emitted.
  */
 #define TSPO_L_ONLY		0x01	/* emit positions appearing only in L */
 #define TSPO_R_ONLY		0x02	/* emit positions appearing only in R */
 #define TSPO_BOTH		0x04	/* emit positions appearing in both L&R */
 
-static TSTernaryValue
+static bool
 TS_phrase_output(ExecPhraseData *data,
 				 ExecPhraseData *Ldata,
 				 ExecPhraseData *Rdata,
@@ -1535,10 +1511,10 @@ TS_phrase_output(ExecPhraseData *data,
 			else
 			{
 				/*
-				 * Exact positions not needed, so return TS_YES as soon as we
+				 * Exact positions not needed, so return true as soon as we
 				 * know there is at least one.
 				 */
-				return TS_YES;
+				return true;
 			}
 		}
 	}
@@ -1547,9 +1523,9 @@ TS_phrase_output(ExecPhraseData *data,
 	{
 		/* Let's assert we didn't overrun the array */
 		Assert(data->npos <= max_npos);
-		return TS_YES;
+		return true;
 	}
-	return TS_NO;
+	return false;
 }
 
 /*
@@ -1567,16 +1543,17 @@ TS_phrase_output(ExecPhraseData *data,
  * This is OK because an outside call always starts from an OP_PHRASE node.
  *
  * The detailed semantics of the match data, given that the function returned
- * TS_YES (successful match), are:
+ * "true" (successful match, or possible match), are:
  *
  * npos > 0, negate = false:
  *	 query is matched at specified position(s) (and only those positions)
  * npos > 0, negate = true:
  *	 query is matched at all positions *except* specified position(s)
+ * npos = 0, negate = false:
+ *	 query is possibly matched, matching position(s) are unknown
+ *	 (this should only be returned when TS_EXEC_PHRASE_NO_POS flag is set)
  * npos = 0, negate = true:
  *	 query is matched at all positions
- * npos = 0, negate = false:
- *	 disallowed (this should result in TS_NO or TS_MAYBE, as appropriate)
  *
  * Successful matches also return a "width" value which is the match width in
  * lexemes, less one.  Hence, "width" is zero for simple one-lexeme matches,
@@ -1585,22 +1562,18 @@ TS_phrase_output(ExecPhraseData *data,
  * the starts.  (This unintuitive rule is needed to avoid possibly generating
  * negative positions, which wouldn't fit into the WordEntryPos arrays.)
  *
- * If the TSExecuteCallback function reports that an operand is present
- * but fails to provide position(s) for it, we will return TS_MAYBE when
- * it is possible but not certain that the query is matched.
- *
- * When the function returns TS_NO or TS_MAYBE, it must return npos = 0,
+ * When the function returns "false" (no match), it must return npos = 0,
  * negate = false (which is the state initialized by the caller); but the
  * "width" output in such cases is undefined.
  */
-static TSTernaryValue
+static bool
 TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
 				  TSExecuteCallback chkcond,
 				  ExecPhraseData *data)
 {
 	ExecPhraseData Ldata,
 				Rdata;
-	TSTernaryValue lmatch,
+	bool		lmatch,
 				rmatch;
 	int			Loffset,
 				Roffset,
@@ -1609,84 +1582,68 @@ TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
 	/* since this function recurses, it could be driven to stack overflow */
 	check_stack_depth();
 
-	/* ... and let's check for query cancel while we're at it */
-	CHECK_FOR_INTERRUPTS();
-
 	if (curitem->type == QI_VAL)
-	{
-		if (!chkcond(arg, (QueryOperand *) curitem, data))
-			return TS_NO;
-		if (data->npos > 0 || data->negate)
-			return TS_YES;
-		/* If we have no position data, we must return TS_MAYBE */
-		return TS_MAYBE;
-	}
+		return chkcond(arg, (QueryOperand *) curitem, data);
 
 	switch (curitem->qoperator.oper)
 	{
 		case OP_NOT:
 
 			/*
-			 * We need not touch data->width, since a NOT operation does not
-			 * change the match width.
+			 * Because a "true" result with no specific positions is taken as
+			 * uncertain, we need no special care here for !TS_EXEC_CALC_NOT.
+			 * If it's a false positive, the right things happen anyway.
+			 *
+			 * Also, we need not touch data->width, since a NOT operation does
+			 * not change the match width.
 			 */
-			if (!(flags & TS_EXEC_CALC_NOT))
+			if (TS_phrase_execute(curitem + 1, arg, flags, chkcond, data))
 			{
-				/* without CALC_NOT, report NOT as "match everywhere" */
+				if (data->npos > 0)
+				{
+					/* we have some positions, invert negate flag */
+					data->negate = !data->negate;
+					return true;
+				}
+				else if (data->negate)
+				{
+					/* change "match everywhere" to "match nowhere" */
+					data->negate = false;
+					return false;
+				}
+				/* match positions are, and remain, uncertain */
+				return true;
+			}
+			else
+			{
+				/* change "match nowhere" to "match everywhere" */
 				Assert(data->npos == 0 && !data->negate);
 				data->negate = true;
-				return TS_YES;
+				return true;
 			}
-			switch (TS_phrase_execute(curitem + 1, arg, flags, chkcond, data))
-			{
-				case TS_NO:
-					/* change "match nowhere" to "match everywhere" */
-					Assert(data->npos == 0 && !data->negate);
-					data->negate = true;
-					return TS_YES;
-				case TS_YES:
-					if (data->npos > 0)
-					{
-						/* we have some positions, invert negate flag */
-						data->negate = !data->negate;
-						return TS_YES;
-					}
-					else if (data->negate)
-					{
-						/* change "match everywhere" to "match nowhere" */
-						data->negate = false;
-						return TS_NO;
-					}
-					/* Should not get here if result was TS_YES */
-					Assert(false);
-					break;
-				case TS_MAYBE:
-					/* match positions are, and remain, uncertain */
-					return TS_MAYBE;
-			}
-			break;
 
 		case OP_PHRASE:
 		case OP_AND:
 			memset(&Ldata, 0, sizeof(Ldata));
 			memset(&Rdata, 0, sizeof(Rdata));
 
-			lmatch = TS_phrase_execute(curitem + curitem->qoperator.left,
-									   arg, flags, chkcond, &Ldata);
-			if (lmatch == TS_NO)
-				return TS_NO;
+			if (!TS_phrase_execute(curitem + curitem->qoperator.left,
+								   arg, flags, chkcond, &Ldata))
+				return false;
 
-			rmatch = TS_phrase_execute(curitem + 1,
-									   arg, flags, chkcond, &Rdata);
-			if (rmatch == TS_NO)
-				return TS_NO;
+			if (!TS_phrase_execute(curitem + 1,
+								   arg, flags, chkcond, &Rdata))
+				return false;
 
 			/*
 			 * If either operand has no position information, then we can't
-			 * return reliable position data, only a MAYBE result.
+			 * return position data, only a "possible match" result. "Possible
+			 * match" answers are only wanted when TS_EXEC_PHRASE_NO_POS flag
+			 * is set, otherwise return false.
 			 */
-			if (lmatch == TS_MAYBE || rmatch == TS_MAYBE)
-				return TS_MAYBE;
+			if ((Ldata.npos == 0 && !Ldata.negate) ||
+				(Rdata.npos == 0 && !Rdata.negate))
+				return (flags & TS_EXEC_PHRASE_NO_POS) ? true : false;
 
 			if (curitem->qoperator.oper == OP_PHRASE)
 			{
@@ -1722,7 +1679,7 @@ TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
 										Ldata.npos + Rdata.npos);
 				if (data)
 					data->negate = true;
-				return TS_YES;
+				return true;
 			}
 			else if (Ldata.negate)
 			{
@@ -1758,24 +1715,27 @@ TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
 			rmatch = TS_phrase_execute(curitem + 1,
 									   arg, flags, chkcond, &Rdata);
 
-			if (lmatch == TS_NO && rmatch == TS_NO)
-				return TS_NO;
+			if (!lmatch && !rmatch)
+				return false;
 
 			/*
-			 * If either operand has no position information, then we can't
-			 * return reliable position data, only a MAYBE result.
+			 * If a valid operand has no position information, then we can't
+			 * return position data, only a "possible match" result. "Possible
+			 * match" answers are only wanted when TS_EXEC_PHRASE_NO_POS flag
+			 * is set, otherwise return false.
 			 */
-			if (lmatch == TS_MAYBE || rmatch == TS_MAYBE)
-				return TS_MAYBE;
+			if ((lmatch && Ldata.npos == 0 && !Ldata.negate) ||
+				(rmatch && Rdata.npos == 0 && !Rdata.negate))
+				return (flags & TS_EXEC_PHRASE_NO_POS) ? true : false;
 
 			/*
 			 * Cope with undefined output width from failed submatch.  (This
 			 * takes less code than trying to ensure that all failure returns
 			 * set data->width to zero.)
 			 */
-			if (lmatch == TS_NO)
+			if (!lmatch)
 				Ldata.width = 0;
-			if (rmatch == TS_NO)
+			if (!rmatch)
 				Rdata.width = 0;
 
 			/*
@@ -1797,7 +1757,7 @@ TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
 										Loffset, Roffset,
 										Min(Ldata.npos, Rdata.npos));
 				data->negate = true;
-				return TS_YES;
+				return true;
 			}
 			else if (Ldata.negate)
 			{
@@ -1807,7 +1767,7 @@ TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
 										Loffset, Roffset,
 										Ldata.npos);
 				data->negate = true;
-				return TS_YES;
+				return true;
 			}
 			else if (Rdata.negate)
 			{
@@ -1817,7 +1777,7 @@ TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
 										Loffset, Roffset,
 										Rdata.npos);
 				data->negate = true;
-				return TS_YES;
+				return true;
 			}
 			else
 			{
@@ -1833,7 +1793,7 @@ TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
 	}
 
 	/* not reachable, but keep compiler quiet */
-	return TS_NO;
+	return false;
 }
 
 
@@ -1844,116 +1804,51 @@ TS_phrase_execute(QueryItem *curitem, void *arg, uint32 flags,
  * arg: opaque value to pass through to callback function
  * flags: bitmask of flag bits shown in ts_utils.h
  * chkcond: callback function to check whether a primitive value is present
+ *
+ * The logic here deals only with operators above any phrase operator, for
+ * which we do not need to worry about lexeme positions.  As soon as we hit an
+ * OP_PHRASE operator, we pass it off to TS_phrase_execute which does worry.
  */
 bool
 TS_execute(QueryItem *curitem, void *arg, uint32 flags,
 		   TSExecuteCallback chkcond)
 {
-	/*
-	 * If we get TS_MAYBE from the recursion, return true.  We could only see
-	 * that result if the caller passed TS_EXEC_PHRASE_NO_POS, so there's no
-	 * need to check again.
-	 */
-	return TS_execute_recurse(curitem, arg, flags, chkcond) != TS_NO;
-}
-
-/*
- * TS_execute recursion for operators above any phrase operator.  Here we do
- * not need to worry about lexeme positions.  As soon as we hit an OP_PHRASE
- * operator, we pass it off to TS_phrase_execute which does worry.
- */
-static TSTernaryValue
-TS_execute_recurse(QueryItem *curitem, void *arg, uint32 flags,
-				   TSExecuteCallback chkcond)
-{
-	TSTernaryValue lmatch;
-
 	/* since this function recurses, it could be driven to stack overflow */
 	check_stack_depth();
 
-	/* ... and let's check for query cancel while we're at it */
-	CHECK_FOR_INTERRUPTS();
-
 	if (curitem->type == QI_VAL)
 		return chkcond(arg, (QueryOperand *) curitem,
-					   NULL /* don't need position info */ ) ? TS_YES : TS_NO;
+					   NULL /* we don't need position info */ );
 
 	switch (curitem->qoperator.oper)
 	{
 		case OP_NOT:
-			if (!(flags & TS_EXEC_CALC_NOT))
-				return TS_YES;
-			switch (TS_execute_recurse(curitem + 1, arg, flags, chkcond))
-			{
-				case TS_NO:
-					return TS_YES;
-				case TS_YES:
-					return TS_NO;
-				case TS_MAYBE:
-					return TS_MAYBE;
-			}
-			break;
+			if (flags & TS_EXEC_CALC_NOT)
+				return !TS_execute(curitem + 1, arg, flags, chkcond);
+			else
+				return true;
 
 		case OP_AND:
-			lmatch = TS_execute_recurse(curitem + curitem->qoperator.left, arg,
-										flags, chkcond);
-			if (lmatch == TS_NO)
-				return TS_NO;
-			switch (TS_execute_recurse(curitem + 1, arg, flags, chkcond))
-			{
-				case TS_NO:
-					return TS_NO;
-				case TS_YES:
-					return lmatch;
-				case TS_MAYBE:
-					return TS_MAYBE;
-			}
-			break;
+			if (TS_execute(curitem + curitem->qoperator.left, arg, flags, chkcond))
+				return TS_execute(curitem + 1, arg, flags, chkcond);
+			else
+				return false;
 
 		case OP_OR:
-			lmatch = TS_execute_recurse(curitem + curitem->qoperator.left, arg,
-										flags, chkcond);
-			if (lmatch == TS_YES)
-				return TS_YES;
-			switch (TS_execute_recurse(curitem + 1, arg, flags, chkcond))
-			{
-				case TS_NO:
-					return lmatch;
-				case TS_YES:
-					return TS_YES;
-				case TS_MAYBE:
-					return TS_MAYBE;
-			}
-			break;
+			if (TS_execute(curitem + curitem->qoperator.left, arg, flags, chkcond))
+				return true;
+			else
+				return TS_execute(curitem + 1, arg, flags, chkcond);
 
 		case OP_PHRASE:
-
-			/*
-			 * If we get a MAYBE result, and the caller doesn't want that,
-			 * convert it to NO.  It would be more consistent, perhaps, to
-			 * return the result of TS_phrase_execute() verbatim and then
-			 * convert MAYBE results at the top of the recursion.  But
-			 * converting at the topmost phrase operator gives results that
-			 * are bug-compatible with the old implementation, so do it like
-			 * this for now.
-			 */
-			switch (TS_phrase_execute(curitem, arg, flags, chkcond, NULL))
-			{
-				case TS_NO:
-					return TS_NO;
-				case TS_YES:
-					return TS_YES;
-				case TS_MAYBE:
-					return (flags & TS_EXEC_PHRASE_NO_POS) ? TS_MAYBE : TS_NO;
-			}
-			break;
+			return TS_phrase_execute(curitem, arg, flags, chkcond, NULL);
 
 		default:
 			elog(ERROR, "unrecognized operator: %d", curitem->qoperator.oper);
 	}
 
 	/* not reachable, but keep compiler quiet */
-	return TS_NO;
+	return false;
 }
 
 /*
@@ -2290,7 +2185,7 @@ ts_setup_firstcall(FunctionCallInfo fcinfo, FuncCallContext *funcctx,
 		}
 	Assert(stat->stackpos <= stat->maxdepth);
 
-	tupdesc = CreateTemplateTupleDesc(3, false);
+	tupdesc = CreateTemplateTupleDesc(3);
 	TupleDescInitEntry(tupdesc, (AttrNumber) 1, "word",
 					   TEXTOID, -1, 0);
 	TupleDescInitEntry(tupdesc, (AttrNumber) 2, "ndoc",
