@@ -2,7 +2,7 @@
  *
  * vacuumdb
  *
- * Portions Copyright (c) 1996-2018, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2019, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * src/bin/scripts/vacuumdb.c
@@ -23,6 +23,8 @@
 #include "catalog/pg_class_d.h"
 
 #include "common.h"
+#include "common/logging.h"
+#include "fe_utils/connect.h"
 #include "fe_utils/simple_list.h"
 #include "fe_utils/string_utils.h"
 
@@ -44,35 +46,40 @@ typedef struct vacuumingOptions
 	bool		and_analyze;
 	bool		full;
 	bool		freeze;
+	bool		disable_page_skipping;
+	bool		skip_locked;
+	int			min_xid_age;
+	int			min_mxid_age;
 } vacuumingOptions;
 
 
-static void vacuum_one_database(const ConnParams *cparams,
-								vacuumingOptions *vacopts,
-					int stage,
-					SimpleStringList *tables,
-					int concurrentCons,
-					const char *progname, bool echo, bool quiet);
+static void vacuum_one_database(const char *dbname, vacuumingOptions *vacopts,
+								int stage,
+								SimpleStringList *tables,
+								const char *host, const char *port,
+								const char *username, enum trivalue prompt_password,
+								int concurrentCons,
+								const char *progname, bool echo, bool quiet);
 
-static void vacuum_all_databases(ConnParams *cparams,
-								 vacuumingOptions *vacopts,
-					 bool analyze_in_stages,
-					 int concurrentCons,
-					 const char *progname, bool echo, bool quiet);
+static void vacuum_all_databases(vacuumingOptions *vacopts,
+								 bool analyze_in_stages,
+								 const char *maintenance_db,
+								 const char *host, const char *port,
+								 const char *username, enum trivalue prompt_password,
+								 int concurrentCons,
+								 const char *progname, bool echo, bool quiet);
 
-static void prepare_vacuum_command(PQExpBuffer sql, PGconn *conn,
-					   vacuumingOptions *vacopts, const char *table,
-					   bool table_pre_qualified,
-					   const char *progname, bool echo);
+static void prepare_vacuum_command(PQExpBuffer sql, int serverVersion,
+								   vacuumingOptions *vacopts, const char *table);
 
 static void run_vacuum_command(PGconn *conn, const char *sql, bool echo,
-				   const char *table, const char *progname, bool async);
+							   const char *table, const char *progname, bool async);
 
 static ParallelSlot *GetIdleSlot(ParallelSlot slots[], int numslots,
-			const char *progname);
+								 const char *progname);
 
 static bool ProcessQueryResult(PGconn *conn, PGresult *result,
-				   const char *progname);
+							   const char *progname);
 
 static bool GetQueryResult(PGconn *conn, const char *progname);
 
@@ -111,6 +118,10 @@ main(int argc, char *argv[])
 		{"jobs", required_argument, NULL, 'j'},
 		{"maintenance-db", required_argument, NULL, 2},
 		{"analyze-in-stages", no_argument, NULL, 3},
+		{"disable-page-skipping", no_argument, NULL, 4},
+		{"skip-locked", no_argument, NULL, 5},
+		{"min-xid-age", required_argument, NULL, 6},
+		{"min-mxid-age", required_argument, NULL, 7},
 		{NULL, 0, NULL, 0}
 	};
 
@@ -123,7 +134,6 @@ main(int argc, char *argv[])
 	char	   *port = NULL;
 	char	   *username = NULL;
 	enum trivalue prompt_password = TRI_DEFAULT;
-	ConnParams	cparams;
 	bool		echo = false;
 	bool		quiet = false;
 	vacuumingOptions vacopts;
@@ -136,8 +146,8 @@ main(int argc, char *argv[])
 	/* initialize options to all false */
 	memset(&vacopts, 0, sizeof(vacopts));
 
+	pg_logging_init(argv[0]);
 	progname = get_progname(argv[0]);
-
 	set_pglocale_pgservice(argv[0], PG_TEXTDOMAIN("pgscripts"));
 
 	handle_help_version_opts(argc, argv, "vacuumdb", help);
@@ -198,8 +208,7 @@ main(int argc, char *argv[])
 				concurrentCons = atoi(optarg);
 				if (concurrentCons <= 0)
 				{
-					fprintf(stderr, _("%s: number of parallel jobs must be at least 1\n"),
-							progname);
+					pg_log_error("number of parallel jobs must be at least 1");
 					exit(1);
 				}
 				break;
@@ -209,6 +218,28 @@ main(int argc, char *argv[])
 			case 3:
 				analyze_in_stages = vacopts.analyze_only = true;
 				break;
+			case 4:
+				vacopts.disable_page_skipping = true;
+				break;
+			case 5:
+				vacopts.skip_locked = true;
+				break;
+			case 6:
+				vacopts.min_xid_age = atoi(optarg);
+				if (vacopts.min_xid_age <= 0)
+				{
+					pg_log_error("minimum transaction ID age must be at least 1");
+					exit(1);
+				}
+				break;
+			case 7:
+				vacopts.min_mxid_age = atoi(optarg);
+				if (vacopts.min_mxid_age <= 0)
+				{
+					pg_log_error("minimum multixact ID age must be at least 1");
+					exit(1);
+				}
+				break;
 			default:
 				fprintf(stderr, _("Try \"%s --help\" for more information.\n"), progname);
 				exit(1);
@@ -227,8 +258,8 @@ main(int argc, char *argv[])
 
 	if (optind < argc)
 	{
-		fprintf(stderr, _("%s: too many command-line arguments (first is \"%s\")\n"),
-				progname, argv[optind]);
+		pg_log_error("too many command-line arguments (first is \"%s\")",
+					 argv[optind]);
 		fprintf(stderr, _("Try \"%s --help\" for more information.\n"), progname);
 		exit(1);
 	}
@@ -237,26 +268,25 @@ main(int argc, char *argv[])
 	{
 		if (vacopts.full)
 		{
-			fprintf(stderr, _("%s: cannot use the \"%s\" option when performing only analyze\n"),
-					progname, "full");
+			pg_log_error("cannot use the \"%s\" option when performing only analyze",
+						 "full");
 			exit(1);
 		}
 		if (vacopts.freeze)
 		{
-			fprintf(stderr, _("%s: cannot use the \"%s\" option when performing only analyze\n"),
-					progname, "freeze");
+			pg_log_error("cannot use the \"%s\" option when performing only analyze",
+						 "freeze");
+			exit(1);
+		}
+		if (vacopts.disable_page_skipping)
+		{
+			pg_log_error("cannot use the \"%s\" option when performing only analyze",
+						 "disable-page-skipping");
 			exit(1);
 		}
 		/* allow 'and_analyze' with 'analyze_only' */
 	}
 
-	/* fill cparams except for dbname, which is set below */
-	cparams.pghost = host;
-	cparams.pgport = port;
-	cparams.pguser = username;
-	cparams.prompt_password = prompt_password;
-	cparams.override_dbname = NULL;
-
 	setup_cancel_handler();
 
 	/* Avoid opening extra connections. */
@@ -267,21 +297,19 @@ main(int argc, char *argv[])
 	{
 		if (dbname)
 		{
-			fprintf(stderr, _("%s: cannot vacuum all databases and a specific one at the same time\n"),
-					progname);
+			pg_log_error("cannot vacuum all databases and a specific one at the same time");
 			exit(1);
 		}
 		if (tables.head != NULL)
 		{
-			fprintf(stderr, _("%s: cannot vacuum specific table(s) in all databases\n"),
-					progname);
+			pg_log_error("cannot vacuum specific table(s) in all databases");
 			exit(1);
 		}
 
-		cparams.dbname = maintenance_db;
-
-		vacuum_all_databases(&cparams, &vacopts,
+		vacuum_all_databases(&vacopts,
 							 analyze_in_stages,
+							 maintenance_db,
+							 host, port, username, prompt_password,
 							 concurrentCons,
 							 progname, echo, quiet);
 	}
@@ -297,25 +325,25 @@ main(int argc, char *argv[])
 				dbname = get_user_name_or_exit(progname);
 		}
 
-		cparams.dbname = dbname;
-
 		if (analyze_in_stages)
 		{
 			int			stage;
 
 			for (stage = 0; stage < ANALYZE_NUM_STAGES; stage++)
 			{
-				vacuum_one_database(&cparams, &vacopts,
+				vacuum_one_database(dbname, &vacopts,
 									stage,
 									&tables,
+									host, port, username, prompt_password,
 									concurrentCons,
 									progname, echo, quiet);
 			}
 		}
 		else
-			vacuum_one_database(&cparams, &vacopts,
+			vacuum_one_database(dbname, &vacopts,
 								ANALYZE_NO_STAGE,
 								&tables,
+								host, port, username, prompt_password,
 								concurrentCons,
 								progname, echo, quiet);
 	}
@@ -337,21 +365,28 @@ main(int argc, char *argv[])
  * a list of tables from the database.
  */
 static void
-vacuum_one_database(const ConnParams *cparams,
-					vacuumingOptions *vacopts,
+vacuum_one_database(const char *dbname, vacuumingOptions *vacopts,
 					int stage,
 					SimpleStringList *tables,
+					const char *host, const char *port,
+					const char *username, enum trivalue prompt_password,
 					int concurrentCons,
 					const char *progname, bool echo, bool quiet)
 {
 	PQExpBufferData sql;
+	PQExpBufferData buf;
+	PQExpBufferData catalog_query;
+	PGresult   *res;
 	PGconn	   *conn;
 	SimpleStringListCell *cell;
 	ParallelSlot *slots;
 	SimpleStringList dbtables = {NULL, NULL};
 	int			i;
+	int			ntups;
 	bool		failed = false;
 	bool		parallel = concurrentCons > 1;
+	bool		tables_listed = false;
+	bool		has_where = false;
 	const char *stage_commands[] = {
 		"SET default_statistics_target=1; SET vacuum_cost_delay=0;",
 		"SET default_statistics_target=10; RESET vacuum_cost_delay;",
@@ -366,7 +401,38 @@ vacuum_one_database(const ConnParams *cparams,
 	Assert(stage == ANALYZE_NO_STAGE ||
 		   (stage >= 0 && stage < ANALYZE_NUM_STAGES));
 
-	conn = connectDatabase(cparams, progname, echo, false, true);
+	conn = connectDatabase(dbname, host, port, username, prompt_password,
+						   progname, echo, false, true);
+
+	if (vacopts->disable_page_skipping && PQserverVersion(conn) < 90600)
+	{
+		PQfinish(conn);
+		pg_log_error("cannot use the \"%s\" option on server versions older than PostgreSQL %s",
+					 "disable-page-skipping", "9.6");
+		exit(1);
+	}
+
+	if (vacopts->skip_locked && PQserverVersion(conn) < 120000)
+	{
+		PQfinish(conn);
+		pg_log_error("cannot use the \"%s\" option on server versions older than PostgreSQL %s",
+					 "skip-locked", "12");
+		exit(1);
+	}
+
+	if (vacopts->min_xid_age != 0 && PQserverVersion(conn) < 90600)
+	{
+		pg_log_error("cannot use the \"%s\" option on server versions older than PostgreSQL %s",
+					 "--min-xid-age", "9.6");
+		exit(1);
+	}
+
+	if (vacopts->min_mxid_age != 0 && PQserverVersion(conn) < 90600)
+	{
+		pg_log_error("cannot use the \"%s\" option on server versions older than PostgreSQL %s",
+					 "--min-mxid-age", "9.6");
+		exit(1);
+	}
 
 	if (!quiet)
 	{
@@ -379,53 +445,175 @@ vacuum_one_database(const ConnParams *cparams,
 		fflush(stdout);
 	}
 
-	initPQExpBuffer(&sql);
-
 	/*
-	 * If a table list is not provided and we're using multiple connections,
-	 * prepare the list of tables by querying the catalogs.
+	 * Prepare the list of tables to process by querying the catalogs.
+	 *
+	 * Since we execute the constructed query with the default search_path
+	 * (which could be unsafe), everything in this query MUST be fully
+	 * qualified.
+	 *
+	 * First, build a WITH clause for the catalog query if any tables were
+	 * specified, with a set of values made of relation names and their
+	 * optional set of columns.  This is used to match any provided column
+	 * lists with the generated qualified identifiers and to filter for the
+	 * tables provided via --table.  If a listed table does not exist, the
+	 * catalog query will fail.
 	 */
-	if (parallel && (!tables || !tables->head))
+	initPQExpBuffer(&catalog_query);
+	for (cell = tables ? tables->head : NULL; cell; cell = cell->next)
 	{
-		PQExpBufferData buf;
-		PGresult   *res;
-		int			ntups;
-
-		initPQExpBuffer(&buf);
-
-		res = executeQuery(conn,
-						   "SELECT c.relname, ns.nspname"
-						   " FROM pg_class c, pg_namespace ns\n"
-						   " WHERE relkind IN ("
-						   CppAsString2(RELKIND_RELATION) ", "
-						   CppAsString2(RELKIND_MATVIEW) ")"
-						   " AND c.relnamespace = ns.oid\n"
-						   " ORDER BY c.relpages DESC;",
-						   progname, echo);
-
-		ntups = PQntuples(res);
-		for (i = 0; i < ntups; i++)
-		{
-			appendPQExpBufferStr(&buf,
-								 fmtQualifiedId(PQgetvalue(res, i, 1),
-												PQgetvalue(res, i, 0)));
-
-			simple_string_list_append(&dbtables, buf.data);
-			resetPQExpBuffer(&buf);
-		}
-
-		termPQExpBuffer(&buf);
-		tables = &dbtables;
+		char	   *just_table;
+		const char *just_columns;
 
 		/*
-		 * If there are more connections than vacuumable relations, we don't
-		 * need to use them all.
+		 * Split relation and column names given by the user, this is used to
+		 * feed the CTE with values on which are performed pre-run validity
+		 * checks as well.  For now these happen only on the relation name.
 		 */
+		splitTableColumnsSpec(cell->val, PQclientEncoding(conn),
+							  &just_table, &just_columns);
+
+		if (!tables_listed)
+		{
+			appendPQExpBuffer(&catalog_query,
+							  "WITH listed_tables (table_oid, column_list) "
+							  "AS (\n  VALUES (");
+			tables_listed = true;
+		}
+		else
+			appendPQExpBuffer(&catalog_query, ",\n  (");
+
+		appendStringLiteralConn(&catalog_query, just_table, conn);
+		appendPQExpBuffer(&catalog_query, "::pg_catalog.regclass, ");
+
+		if (just_columns && just_columns[0] != '\0')
+			appendStringLiteralConn(&catalog_query, just_columns, conn);
+		else
+			appendPQExpBufferStr(&catalog_query, "NULL");
+
+		appendPQExpBufferStr(&catalog_query, "::pg_catalog.text)");
+
+		pg_free(just_table);
+	}
+
+	/* Finish formatting the CTE */
+	if (tables_listed)
+		appendPQExpBuffer(&catalog_query, "\n)\n");
+
+	appendPQExpBuffer(&catalog_query, "SELECT c.relname, ns.nspname");
+
+	if (tables_listed)
+		appendPQExpBuffer(&catalog_query, ", listed_tables.column_list");
+
+	appendPQExpBuffer(&catalog_query,
+					  " FROM pg_catalog.pg_class c\n"
+					  " JOIN pg_catalog.pg_namespace ns"
+					  " ON c.relnamespace OPERATOR(pg_catalog.=) ns.oid\n"
+					  " LEFT JOIN pg_catalog.pg_class t"
+					  " ON c.reltoastrelid OPERATOR(pg_catalog.=) t.oid\n");
+
+	/* Used to match the tables listed by the user */
+	if (tables_listed)
+		appendPQExpBuffer(&catalog_query, " JOIN listed_tables"
+						  " ON listed_tables.table_oid OPERATOR(pg_catalog.=) c.oid\n");
+
+	/*
+	 * If no tables were listed, filter for the relevant relation types.  If
+	 * tables were given via --table, don't bother filtering by relation type.
+	 * Instead, let the server decide whether a given relation can be
+	 * processed in which case the user will know about it.
+	 */
+	if (!tables_listed)
+	{
+		appendPQExpBuffer(&catalog_query, " WHERE c.relkind OPERATOR(pg_catalog.=) ANY (array["
+						  CppAsString2(RELKIND_RELATION) ", "
+						  CppAsString2(RELKIND_MATVIEW) "])\n");
+		has_where = true;
+	}
+
+	/*
+	 * For --min-xid-age and --min-mxid-age, the age of the relation is the
+	 * greatest of the ages of the main relation and its associated TOAST
+	 * table.  The commands generated by vacuumdb will also process the TOAST
+	 * table for the relation if necessary, so it does not need to be
+	 * considered separately.
+	 */
+	if (vacopts->min_xid_age != 0)
+	{
+		appendPQExpBuffer(&catalog_query,
+						  " %s GREATEST(pg_catalog.age(c.relfrozenxid),"
+						  " pg_catalog.age(t.relfrozenxid)) "
+						  " OPERATOR(pg_catalog.>=) '%d'::pg_catalog.int4\n"
+						  " AND c.relfrozenxid OPERATOR(pg_catalog.!=)"
+						  " '0'::pg_catalog.xid\n",
+						  has_where ? "AND" : "WHERE", vacopts->min_xid_age);
+		has_where = true;
+	}
+
+	if (vacopts->min_mxid_age != 0)
+	{
+		appendPQExpBuffer(&catalog_query,
+						  " %s GREATEST(pg_catalog.mxid_age(c.relminmxid),"
+						  " pg_catalog.mxid_age(t.relminmxid)) OPERATOR(pg_catalog.>=)"
+						  " '%d'::pg_catalog.int4\n"
+						  " AND c.relminmxid OPERATOR(pg_catalog.!=)"
+						  " '0'::pg_catalog.xid\n",
+						  has_where ? "AND" : "WHERE", vacopts->min_mxid_age);
+		has_where = true;
+	}
+
+	/*
+	 * Execute the catalog query.  We use the default search_path for this
+	 * query for consistency with table lookups done elsewhere by the user.
+	 */
+	appendPQExpBuffer(&catalog_query, " ORDER BY c.relpages DESC;");
+	executeCommand(conn, "RESET search_path;", progname, echo);
+	res = executeQuery(conn, catalog_query.data, progname, echo);
+	termPQExpBuffer(&catalog_query);
+	PQclear(executeQuery(conn, ALWAYS_SECURE_SEARCH_PATH_SQL,
+						 progname, echo));
+
+	/*
+	 * If no rows are returned, there are no matching tables, so we are done.
+	 */
+	ntups = PQntuples(res);
+	if (ntups == 0)
+	{
+		PQclear(res);
+		PQfinish(conn);
+		return;
+	}
+
+	/*
+	 * Build qualified identifiers for each table, including the column list
+	 * if given.
+	 */
+	initPQExpBuffer(&buf);
+	for (i = 0; i < ntups; i++)
+	{
+		appendPQExpBufferStr(&buf,
+							 fmtQualifiedId(PQgetvalue(res, i, 1),
+											PQgetvalue(res, i, 0)));
+
+		if (tables_listed && !PQgetisnull(res, i, 2))
+			appendPQExpBufferStr(&buf, PQgetvalue(res, i, 2));
+
+		simple_string_list_append(&dbtables, buf.data);
+		resetPQExpBuffer(&buf);
+	}
+	termPQExpBuffer(&buf);
+	PQclear(res);
+
+	/*
+	 * If there are more connections than vacuumable relations, we don't need
+	 * to use them all.
+	 */
+	if (parallel)
+	{
 		if (concurrentCons > ntups)
 			concurrentCons = ntups;
 		if (concurrentCons <= 1)
 			parallel = false;
-		PQclear(res);
 	}
 
 	/*
@@ -441,7 +629,8 @@ vacuum_one_database(const ConnParams *cparams,
 	{
 		for (i = 1; i < concurrentCons; i++)
 		{
-			conn = connectDatabase(cparams, progname, echo, false, true);
+			conn = connectDatabase(dbname, host, port, username, prompt_password,
+								   progname, echo, false, true);
 
 			/*
 			 * Fail and exit immediately if trying to use a socket in an
@@ -450,9 +639,7 @@ vacuum_one_database(const ConnParams *cparams,
 			 */
 			if (PQsocket(conn) >= FD_SETSIZE)
 			{
-				fprintf(stderr,
-						_("%s: too many jobs for this platform -- try %d\n"),
-						progname, i);
+				pg_log_fatal("too many jobs for this platform -- try %d", i);
 				exit(1);
 			}
 
@@ -475,10 +662,12 @@ vacuum_one_database(const ConnParams *cparams,
 						   stage_commands[stage], progname, echo);
 	}
 
-	cell = tables ? tables->head : NULL;
+	initPQExpBuffer(&sql);
+
+	cell = dbtables.head;
 	do
 	{
-		const char *tabname = cell ? cell->val : NULL;
+		const char *tabname = cell->val;
 		ParallelSlot *free_slot;
 
 		if (CancelRequested)
@@ -511,12 +700,8 @@ vacuum_one_database(const ConnParams *cparams,
 		else
 			free_slot = slots;
 
-		/*
-		 * Prepare the vacuum command.  Note that in some cases this requires
-		 * query execution, so be sure to use the free connection.
-		 */
-		prepare_vacuum_command(&sql, free_slot->connection, vacopts, tabname,
-							   tables == &dbtables, progname, echo);
+		prepare_vacuum_command(&sql, PQserverVersion(free_slot->connection),
+							   vacopts, tabname);
 
 		/*
 		 * Execute the vacuum.  If not in parallel mode, this terminates the
@@ -526,8 +711,7 @@ vacuum_one_database(const ConnParams *cparams,
 		run_vacuum_command(free_slot->connection, sql.data,
 						   echo, tabname, progname, parallel);
 
-		if (cell)
-			cell = cell->next;
+		cell = cell->next;
 	} while (cell != NULL);
 
 	if (parallel)
@@ -564,23 +748,28 @@ finish:
  * quickly everywhere before generating more detailed ones.
  */
 static void
-vacuum_all_databases(ConnParams *cparams,
-					 vacuumingOptions *vacopts,
+vacuum_all_databases(vacuumingOptions *vacopts,
 					 bool analyze_in_stages,
+					 const char *maintenance_db, const char *host,
+					 const char *port, const char *username,
+					 enum trivalue prompt_password,
 					 int concurrentCons,
 					 const char *progname, bool echo, bool quiet)
 {
 	PGconn	   *conn;
 	PGresult   *result;
+	PQExpBufferData connstr;
 	int			stage;
 	int			i;
 
-	conn = connectMaintenanceDatabase(cparams, progname, echo);
+	conn = connectMaintenanceDatabase(maintenance_db, host, port, username,
+									  prompt_password, progname, echo);
 	result = executeQuery(conn,
-						  "SELECT datname FROM pg_database WHERE datallowconn AND datconnlimit <> -2 ORDER BY 1;",
+						  "SELECT datname FROM pg_database WHERE datallowconn ORDER BY 1;",
 						  progname, echo);
 	PQfinish(conn);
 
+	initPQExpBuffer(&connstr);
 	if (analyze_in_stages)
 	{
 		/*
@@ -595,11 +784,14 @@ vacuum_all_databases(ConnParams *cparams,
 		{
 			for (i = 0; i < PQntuples(result); i++)
 			{
-				cparams->override_dbname = PQgetvalue(result, i, 0);
+				resetPQExpBuffer(&connstr);
+				appendPQExpBuffer(&connstr, "dbname=");
+				appendConnStrVal(&connstr, PQgetvalue(result, i, 0));
 
-				vacuum_one_database(cparams, vacopts,
+				vacuum_one_database(connstr.data, vacopts,
 									stage,
 									NULL,
+									host, port, username, prompt_password,
 									concurrentCons,
 									progname, echo, quiet);
 			}
@@ -609,15 +801,19 @@ vacuum_all_databases(ConnParams *cparams,
 	{
 		for (i = 0; i < PQntuples(result); i++)
 		{
-			cparams->override_dbname = PQgetvalue(result, i, 0);
+			resetPQExpBuffer(&connstr);
+			appendPQExpBuffer(&connstr, "dbname=");
+			appendConnStrVal(&connstr, PQgetvalue(result, i, 0));
 
-			vacuum_one_database(cparams, vacopts,
+			vacuum_one_database(connstr.data, vacopts,
 								ANALYZE_NO_STAGE,
 								NULL,
+								host, port, username, prompt_password,
 								concurrentCons,
 								progname, echo, quiet);
 		}
 	}
+	termPQExpBuffer(&connstr);
 
 	PQclear(result);
 }
@@ -626,32 +822,68 @@ vacuum_all_databases(ConnParams *cparams,
  * Construct a vacuum/analyze command to run based on the given options, in the
  * given string buffer, which may contain previous garbage.
  *
- * An optional table name can be passed; this must be already be properly
- * quoted.  The command is semicolon-terminated.
+ * The table name used must be already properly quoted.  The command generated
+ * depends on the server version involved and it is semicolon-terminated.
  */
 static void
-prepare_vacuum_command(PQExpBuffer sql, PGconn *conn,
-					   vacuumingOptions *vacopts, const char *table,
-					   bool table_pre_qualified,
-					   const char *progname, bool echo)
+prepare_vacuum_command(PQExpBuffer sql, int serverVersion,
+					   vacuumingOptions *vacopts, const char *table)
 {
+	const char *paren = " (";
+	const char *comma = ", ";
+	const char *sep = paren;
+
 	resetPQExpBuffer(sql);
 
 	if (vacopts->analyze_only)
 	{
 		appendPQExpBufferStr(sql, "ANALYZE");
-		if (vacopts->verbose)
-			appendPQExpBufferStr(sql, " VERBOSE");
+
+		/* parenthesized grammar of ANALYZE is supported since v11 */
+		if (serverVersion >= 110000)
+		{
+			if (vacopts->skip_locked)
+			{
+				/* SKIP_LOCKED is supported since v12 */
+				Assert(serverVersion >= 120000);
+				appendPQExpBuffer(sql, "%sSKIP_LOCKED", sep);
+				sep = comma;
+			}
+			if (vacopts->verbose)
+			{
+				appendPQExpBuffer(sql, "%sVERBOSE", sep);
+				sep = comma;
+			}
+			if (sep != paren)
+				appendPQExpBufferChar(sql, ')');
+		}
+		else
+		{
+			if (vacopts->verbose)
+				appendPQExpBufferStr(sql, " VERBOSE");
+		}
 	}
 	else
 	{
 		appendPQExpBufferStr(sql, "VACUUM");
-		if (PQserverVersion(conn) >= 90000)
-		{
-			const char *paren = " (";
-			const char *comma = ", ";
-			const char *sep = paren;
 
+		/* parenthesized grammar of VACUUM is supported since v9.0 */
+		if (serverVersion >= 90000)
+		{
+			if (vacopts->disable_page_skipping)
+			{
+				/* DISABLE_PAGE_SKIPPING is supported since v9.6 */
+				Assert(serverVersion >= 90600);
+				appendPQExpBuffer(sql, "%sDISABLE_PAGE_SKIPPING", sep);
+				sep = comma;
+			}
+			if (vacopts->skip_locked)
+			{
+				/* SKIP_LOCKED is supported since v12 */
+				Assert(serverVersion >= 120000);
+				appendPQExpBuffer(sql, "%sSKIP_LOCKED", sep);
+				sep = comma;
+			}
 			if (vacopts->full)
 			{
 				appendPQExpBuffer(sql, "%sFULL", sep);
@@ -688,15 +920,7 @@ prepare_vacuum_command(PQExpBuffer sql, PGconn *conn,
 		}
 	}
 
-	if (table)
-	{
-		appendPQExpBufferChar(sql, ' ');
-		if (table_pre_qualified)
-			appendPQExpBufferStr(sql, table);
-		else
-			appendQualifiedRelation(sql, table, conn, progname, echo);
-	}
-	appendPQExpBufferChar(sql, ';');
+	appendPQExpBuffer(sql, " %s;", table);
 }
 
 /*
@@ -725,12 +949,11 @@ run_vacuum_command(PGconn *conn, const char *sql, bool echo,
 	if (!status)
 	{
 		if (table)
-			fprintf(stderr,
-					_("%s: vacuuming of table \"%s\" in database \"%s\" failed: %s"),
-					progname, table, PQdb(conn), PQerrorMessage(conn));
+			pg_log_error("vacuuming of table \"%s\" in database \"%s\" failed: %s",
+						 table, PQdb(conn), PQerrorMessage(conn));
 		else
-			fprintf(stderr, _("%s: vacuuming of database \"%s\" failed: %s"),
-					progname, PQdb(conn), PQerrorMessage(conn));
+			pg_log_error("vacuuming of database \"%s\" failed: %s",
+						 PQdb(conn), PQerrorMessage(conn));
 
 		if (!async)
 		{
@@ -864,8 +1087,8 @@ ProcessQueryResult(PGconn *conn, PGresult *result, const char *progname)
 	{
 		char	   *sqlState = PQresultErrorField(result, PG_DIAG_SQLSTATE);
 
-		fprintf(stderr, _("%s: vacuuming of database \"%s\" failed: %s"),
-				progname, PQdb(conn), PQerrorMessage(conn));
+		pg_log_error("vacuuming of database \"%s\" failed: %s",
+					 PQdb(conn), PQerrorMessage(conn));
 
 		if (sqlState && strcmp(sqlState, ERRCODE_UNDEFINED_TABLE) != 0)
 		{
@@ -1005,11 +1228,15 @@ help(const char *progname)
 	printf(_("\nOptions:\n"));
 	printf(_("  -a, --all                       vacuum all databases\n"));
 	printf(_("  -d, --dbname=DBNAME             database to vacuum\n"));
+	printf(_("      --disable-page-skipping     disable all page-skipping behavior\n"));
 	printf(_("  -e, --echo                      show the commands being sent to the server\n"));
 	printf(_("  -f, --full                      do full vacuuming\n"));
 	printf(_("  -F, --freeze                    freeze row transaction information\n"));
 	printf(_("  -j, --jobs=NUM                  use this many concurrent connections to vacuum\n"));
+	printf(_("      --min-mxid-age=MXID_AGE     minimum multixact ID age of tables to vacuum\n"));
+	printf(_("      --min-xid-age=XID_AGE       minimum transaction ID age of tables to vacuum\n"));
 	printf(_("  -q, --quiet                     don't write any messages\n"));
+	printf(_("      --skip-locked               skip relations that cannot be immediately locked\n"));
 	printf(_("  -t, --table='TABLE[(COLUMNS)]'  vacuum specific table(s) only\n"));
 	printf(_("  -v, --verbose                   write a lot of output\n"));
 	printf(_("  -V, --version                   output version information, then exit\n"));
@@ -1026,5 +1253,5 @@ help(const char *progname)
 	printf(_("  -W, --password            force password prompt\n"));
 	printf(_("  --maintenance-db=DBNAME   alternate maintenance database\n"));
 	printf(_("\nRead the description of the SQL command VACUUM for details.\n"));
-	printf(_("\nReport bugs to <pgsql-bugs@postgresql.org>.\n"));
+	printf(_("\nReport bugs to <pgsql-bugs@lists.postgresql.org>.\n"));
 }
